#include "Displayer.h"

#include "opencv2/highgui/highgui.hpp"
#include "opencv2/imgproc/imgproc.hpp"

/**
 * Displays the two given images side by side on the screen
 */
void Displayer::displayTwoImgs(const Mat& img1, const Mat& img2) {
	Mat& correspond = mergeTwoImgs(img1, img2);
	displayMat(correspond);
}

/**
 * Displays two images side by side and all their keypoints
 */
void Displayer::displayKeypoints(const Mat& img1, const std::vector<KeyPoint>& img1Keypoints,
								 const Mat& img2, const std::vector<KeyPoint>& img2Keypoints, int color) {
	Mat keypointsImg1, keypointsImg2;
	drawKeypoints(img1, img1Keypoints, keypointsImg1, color);
	drawKeypoints(img2, img2Keypoints, keypointsImg2, color);

	Mat& correspond = mergeTwoImgs(keypointsImg1, keypointsImg2);

	displayMat(correspond);
}

/**
 * Given the vector of matches, displays the two images side by side
 * and draws lines between all the matches
 */
void Displayer::displayMatches(const Mat& img1, const Mat& img2, const vector<Match>& matches, int color) {
	Mat& correspond = mergeTwoImgs(img1, img2);
	CvScalar c = {{color, color, color}};

	for (size_t i = 0; i < matches.size(); ++i) {
		displayMatch(correspond, matches[i], img1.cols, c);
	}
	displayMat(correspond);
}

/**
 * Given the vector of matches, displays the two images side by side
 * and draws lines between all the matches
 *
 * This is is the same as above except uses the built in drawMatches function
 */
void Displayer::displayMatches(const Mat& img1, const std::vector<KeyPoint>& img1Keypoints,
							   const Mat& img2, const std::vector<KeyPoint>& img2Keypoints,
							   const vector<DMatch>& matches) {

	Mat correspond;
	drawMatches(img1, img1Keypoints, img2, img2Keypoints, matches, correspond);
	displayMat(correspond);
}

/**
 * Given the vector of matches, displays the two image side by side
 * and draws lines between the matches which are not considered outliers
 */
void Displayer::displayMatches(const Mat& img1, const Mat& img2, const vector<Match>& matches,
							   const std::vector<double>& supportTable, double median, double medianScale, int color) {
	Mat& correspond = mergeTwoImgs(img1, img2);
	CvScalar c = {{color, color, color}};

	for (size_t i = 0; i < matches.size(); ++i) {
		if (supportTable[i] < median * medianScale) {			
			displayMatch(correspond, matches[i], img1.cols, c);
		}
	}
	displayMat(correspond);
}

/**
 * Given two sets of MSER contours, displays them side by side
 * The original images are passed only for dimension information
 */
void Displayer::displayMSERs(const Mat& img1, const Mat& img2, const std::vector<std::vector<Point>>& msers1,
							 const std::vector<std::vector<Point>>& msers2) {
	RNG rng(12345);
	vector<Vec4i> hierarchy;
	Mat contours1 = Mat::zeros( img1.size(), CV_8UC1 );
	Mat contours2 = Mat::zeros( img2.size(), CV_8UC1 );

	for( size_t i = 0; i< msers1.size(); i++ )
	{
		// Draw the contour
		Scalar color = Scalar( rng.uniform(100, 255), rng.uniform(100,255), rng.uniform(100,255) );	// random colour
		drawContours( contours1, msers1, i, color, 2, 8, hierarchy, 0, Point() );
	}

	for( size_t i = 0; i< msers2.size(); i++ )
	{
		// Draw the contour
		Scalar color = Scalar( rng.uniform(100, 255), rng.uniform(100,255), rng.uniform(100,255) );	// random colour
		drawContours( contours2, msers2, i, color, 2, 8, hierarchy, 0, Point() );
	}
	
	Mat& correspond = mergeTwoImgs(contours1, contours2);
	displayMat(correspond);
}

/**
 * Helper function to draw a match on an image
 */
void Displayer::displayMatch(Mat& correspond, const Match& match, int firstImageWidth, CvScalar color) {
	Point2f p1 = match.p1;
	Point2f p2 = Point2f(match.p2.x + firstImageWidth, match.p2.y);		// offset by width of img1
	circle(correspond, p1, 4, color);
	circle(correspond, p2, 4, color);
	line(correspond, p1, p2, color);
}

/**
 * Takes two Mat objects and merges them into large one
 */
Mat Displayer::mergeTwoImgs(const Mat& img1, const Mat& img2) {
	Mat correspond(max(img1.rows, img2.rows), img1.cols + img2.cols, CV_8UC1);
	// Create a region of interest in the correspond Matrix
	// Its a reference and any changes to it are done to correspond
	Mat roi(correspond, Rect(0, 0, img1.cols, img1.rows));
	img1.copyTo(roi);
	roi = Mat(correspond, Rect(img1.cols, 0, img2.cols, img2.rows));
	img2.copyTo(roi);
	return correspond;
}

/**
 * For some reason the OpenCV drawKeypoints output wouldn't work with mergeTwoImgs()
 * so implement it explicitly
 */
void Displayer::drawKeypoints(const Mat& img1, const std::vector<KeyPoint>& keypoints, Mat& outputImage, int color) {
	outputImage = img1.clone();
	CvScalar c = {{color, color, color}};

	for (size_t i = 0; i < keypoints.size(); ++i) {
		circle(outputImage, keypoints.at(i).pt, 4, c);
	}
}

/**
 * Displays the given image
 */
void Displayer::displayMat(const Mat& img) {
	imshow("Correspond", img);
}