#include "FeatureSystem.h"
#include "NeighbourhoodDescriptorExtractor.h"
#include <iostream>

FeatureSystem::FeatureSystem() : disparityMedianScale(2.0) {}

std::string FeatureSystem::getSystemName(int type) {
	return getSystemName(static_cast<Type>(type));
}

/**
 * Return the string representation of a feature system given its type
 */
std::string FeatureSystem::getSystemName(Type type) {
	switch(type) {
	case HARRIS:
		return "HARRIS";
	case KLT:
		return "KLT";
	case SIFT:
		return "SIFT";
	case SURF:
		return "SURF";
	}
	return "";
}

/**
 * Executes FLANN on the two images, given their keypoints and descriptors
 *
 * @returns a vector of the keypoints that match
 */
void FeatureSystem::flannFindMatches(const std::vector<KeyPoint>& img1Keypoints, const Mat& img1Descriptors,
									 const std::vector<KeyPoint>& img2Keypoints, const Mat& img2Descriptors,
									 std::vector<Match>& matches) {
	// find nearest neighbors using FLANN
	cv::Mat m_indices(img1Descriptors.total(), 2, CV_32S);
	cv::Mat m_dists(img1Descriptors.total(), 2, CV_32F);
	cv::flann::Index flann_index(img2Descriptors, cv::flann::KDTreeIndexParams(4));  // using 4 randomized kdtrees
	flann_index.knnSearch(img1Descriptors, m_indices, m_dists, 2, cv::flann::SearchParams(64) ); // maximum number of leafs checked

	int* indices_ptr = m_indices.ptr<int>(0);
	float* dists_ptr = m_dists.ptr<float>(0);
	for (int i = 0; i < m_indices.rows; ++i) {
		// If the two keypoints are similar enough
		if (dists_ptr[2 * i] < 0.6 * dists_ptr[2 * i + 1]) {
			matches.push_back(Match(img1Keypoints[i].pt,
									img2Keypoints[indices_ptr[i * 2]].pt));
		}
	}
}

/**
 * Used for the disparity filter, code came from a OpenCV example
 * Finds the kth value in the given vector from indices 1-right
 * Corrupts the given error_vector, so it shouldn't be used after calling this
 */
double FeatureSystem::find_kth(long int k, std::vector<double>& error_vector, long int right) {
	double value, temp;
	long int left, i, j;

	left = 1;  
	error_vector[0] = VERY_LARGE_NEG_NUM;

	while (right > left)
	{
		value = error_vector[right];
		i = left - 1;
		j = right;
		do {
			do ++i; while (i<right && error_vector[i] < value);
			do --j; while (j>0 && error_vector[j] > value);
			temp = error_vector[i];
			error_vector[i] = error_vector[j];
			error_vector[j] = temp;
		} while (j > i);

		error_vector[j] = error_vector[i];
		error_vector[i] = error_vector[right];
		error_vector[right] = temp;

		if (i >= k) 
			right = i - 1;
		if (i <= k) 
			left = i + 1;
	}

	return (error_vector[k + 1]);
}

/**
 * Returns the disparity gradient of two nearby points.
 * It is the difference in their disparities divided by their separation in visual angle
 */
double FeatureSystem::disparity_gradient(Point2f from1, Point2f to1, Point2f from2, Point2f to2) {
	double visualAngleSeparation, disparityDifference;
	Point2f point1, point2;

	/* cyclopean seperation is vector between midpoints of gradients */
	point1.x = (from1.x + to1.x) / 2;
	point1.y = (from1.y + to1.y) / 2;
	
	point2.x = (from2.x + to2.x) / 2;
	point2.y = (from2.y + to2.y) / 2;

	visualAngleSeparation = norm(point2 - point1);		// Length of vector formed by point1 to point 2
 
	/* difference vector of disparities */
	point1.x = to1.x - from1.x;
	point1.y = to1.y - from1.y;

	point2.x = to2.x - from2.x;
	point2.y = to2.y - from2.y;

	disparityDifference = norm(point2 - point1);		// Length of vector formed by point1 to point 2

	return disparityDifference / visualAngleSeparation;
}

/**
 * Given a vector of matches, returns a vector of the disparity gradient, as well as the median value
 *
 * The disparity gradient is used to locate outliers in the matches, which can then be discarded
 * A disparity gradient value is given to each match, and if the value is greater than the median times the medianScale,
 * it is considered an outlier
 */
void FeatureSystem::disparityGradientFilter(const std::vector<Match>& matches, std::vector<double>& dispgrad, double *median) {
	size_t i, j;
	const size_t correspondence_count = matches.size();
	bool one_corresp_deleted = false;
	int current_count;
	
	std::vector<double> temp_vector(correspondence_count + 2);
	std::vector<bool> validPoints(correspondence_count + 1);
	std::vector<std::vector<double>> alldispgrads(correspondence_count);

	for (i = 0; i < correspondence_count; ++i) {
		validPoints[i] = true;
		alldispgrads[i].assign(correspondence_count, 0.0);
	}

	/* COMPUTE THE DISP GRADS ONLY ONCE */
	for (i = 0; i < correspondence_count; ++i)
		for (j = 0; j <=i; j++) {
			if (i == j) 
				alldispgrads[i][j] = 0.0;
			else {
				alldispgrads[i][j] += disparity_gradient(matches[i].p1, matches[i].p2,
														 matches[j].p1, matches[j].p2);
			}
			alldispgrads[j][i] = alldispgrads[i][j];
		}
	do {
		one_corresp_deleted = 0;
		dispgrad.assign(correspondence_count + 1, 0);		// Initialize to zero
		
		for (i = 0; i < correspondence_count; i++)
			for (j = 0; j < correspondence_count; j++)
				if (i != j && validPoints[i+1] && validPoints[j+1])
					dispgrad[i+1] += alldispgrads[i][j];

		current_count = 1;
		// since the find_kth destroys the vector...
		for (i = 0; i <= correspondence_count; i++) 
			if(validPoints[i])
				temp_vector[current_count++] = dispgrad[i];

		--current_count;
		*median = find_kth(current_count/2, temp_vector, current_count);

		/* if median is too small everyone is an inlier */
		if (fabs(*median) < VERY_SMALL_POS_NUM)
		  *median = VERY_LARGE_POS_NUM;

		for (i = 1; i <= correspondence_count; i++) {
			if(validPoints[i])
				if (dispgrad[i] > *median * disparityMedianScale) {
					validPoints[i] = false;
					one_corresp_deleted = true;
				}
		}
	} while (one_corresp_deleted);

	/* MARK OUR INVALID POINTS VIA DISPGRAD ARRAY */
	for (i = 1; i <= correspondence_count; i++) {
		if(!validPoints[i]) {
			dispgrad[i-1] = 999999;
		}
	}
}

/**
 * Predicate used for removeEdgeKeypoints()
 * a keypoint is an edge if its neighbourhood goes outside the image
 */
struct IsEdgeKeypoint {
	IsEdgeKeypoint(const cv::Mat& image) : _image(image) {}

	bool operator()(const cv::KeyPoint v) {
		// Return true if keypoint is at edges of image
		const int edgeStart = NeighbourhoodDescriptorExtractor::HALF_SIZE;
		cv::Point pt((int)v.pt.x, (int) v.pt.y);		// Convert location to ints

		bool result =	pt.x < edgeStart || pt.x >= _image.cols - edgeStart ||
						pt.y < edgeStart || pt.y >= _image.rows - edgeStart;

		if (result) {
			// Print out which keypoints were removed
			std::cout << "DEBUG: removing keypoint: " << pt << std::endl;
		}

		return result;
	}

	const cv::Mat& _image;
};

/**
 * For a given image and its vector of keypoints, removes any keypoints that are considered past the edge
 */
void FeatureSystem::removeEdgeKeypoints(const cv::Mat& image, std::vector<cv::KeyPoint>& keypoints) {
	IsEdgeKeypoint predicate(image);

	keypoints.erase(
		// Moves all the edge keypoints to the end of the container
		// Returns an iterator to the element that follows the last element not removed
		std::remove_if(keypoints.begin(),
					   keypoints.end(),
					   predicate),
		keypoints.end()
		// This erases all the items at the end of the container
	);
}