#ifndef FEATURE_SYSTEM_H
#define FEATURE_SYSTEM_H

#include "opencv2/core/core.hpp"
#include "opencv2/features2d/features2d.hpp"

using namespace cv;

// Used for disparity gradient
#define VERY_LARGE_POS_NUM              +1e30
#define VERY_LARGE_NEG_NUM              -1e30
#define VERY_SMALL_POS_NUM              +1e-20
#define VERY_SMALL_NEG_NUM              -1e-20

/**
 * Represents a match between two keypoints
 */
struct Match {
	Match(Point2f p1_, Point2f p2_) : p1(p1_), p2(p2_) {}
	Point2f p1, p2;
};

/**
 * Abstract base class for the feature systems
 */
class FeatureSystem {
public:
	enum Type {
		HARRIS	= 0,
		KLT		= 1,
		SIFT	= 2,
		SURF	= 3,
		COUNT	= 4		// 'size' of Type
	};

	const double disparityMedianScale;		// How much to multiply the median by in the disparity filter

	FeatureSystem();

	// Feature system subclass must implement the following
	virtual void findFeatures(const cv::Mat& image, std::vector<cv::KeyPoint>& keypoints) = 0;
	virtual void findDescriptors(const cv::Mat& image, std::vector<cv::KeyPoint>& keypoints, Mat& descriptors) = 0;

	void flannFindMatches(const std::vector<KeyPoint>& img1Keypoints, const Mat& img1Descriptors,
						  const std::vector<KeyPoint>& img2Keypoints, const Mat& img2Descriptors,
						  std::vector<Match>& matches);

	void disparityGradientFilter(const std::vector<Match>& matches, std::vector<double>& dispgrad, double *median);

	static std::string getSystemName(int type);
	static std::string getSystemName(Type type);

protected:	
	double find_kth(long int k, std::vector<double>& error_vector, long int right);
	double disparity_gradient(Point2f from1, Point2f to1, Point2f from2, Point2f to2);
	void removeEdgeKeypoints(const cv::Mat& image, std::vector<cv::KeyPoint>& keypoints);
};

#endif