#include "NeighbourhoodDescriptorExtractor.h"
#include <limits>
#include <algorithm>

const int NeighbourhoodDescriptorExtractor::NEIGHBOURHOOD_SIZE = 7;
const int NeighbourhoodDescriptorExtractor::HALF_SIZE = NEIGHBOURHOOD_SIZE / 2;

/**
 * For a given image and its keypoints, outputs a Matrix of the descriptors of each keypoint
 * This type of descriptor is the K pixel neighbourhood around the keypoint (inclusive) and their intensity values
 * K can be any odd number, 3, 7, 9.
 *
 * The output matrix has a row for each keypoint, and each column is a pixel in the neighbourhood.
 * They neighbourhood pixels for each row are ordered as follows (using a 3x3 neighbourhood as an example):
 *
 *		1 2 3
 *		4 5 6
 *		7 8 9
 *
 */
void NeighbourhoodDescriptorExtractor::extract(const Mat& image, const vector<KeyPoint>& keyPoints, Mat& outDescriptors) {
	// The pixels of the image and the # of channels
	unsigned char* pixelPtr = (unsigned char*)image.data;
	int cn = image.channels();

	// Loop through each keypoint
	vector<KeyPoint>::const_iterator iter;

	// A row for each descriptor, neighbourhood values
	const int NUM_NEIGHBOURS = NEIGHBOURHOOD_SIZE * NEIGHBOURHOOD_SIZE;
	outDescriptors = Mat(keyPoints.size(), NUM_NEIGHBOURS, CV_32F);		// Descriptors must be CV_32F for FLANN
	int currentRow = 0;
	float* descriptorPtr = (float*) outDescriptors.data;

	for (iter = keyPoints.begin(); iter != keyPoints.end(); ++iter) {
		// Get the pixel neighbourhood of the keypoint
		unsigned char max = std::numeric_limits<unsigned char>::min();
		unsigned char min = std::numeric_limits<unsigned char>::max();

		int x = (int)iter->pt.x, y = (int)iter->pt.y;
		int currentColumn = 0;

		// First get the max and min of the neighbourhood
		for (int i = - HALF_SIZE; i < HALF_SIZE + 1; ++i) {
			for (int j = - HALF_SIZE; j < HALF_SIZE + 1; ++j) {
				int newX = x + j;
				int newY = y + i;

				unsigned char pixelValue = pixelPtr[newY * image.cols * cn + newX * cn];
				max = std::max(pixelValue, max);
				min = std::min(pixelValue, min);
			}
		}
		// Normalize each value by making it (Value - MIN) / (MAX - MIN)
		float range = (float) (max - min);
		for (int i = - HALF_SIZE; i < HALF_SIZE + 1; ++i) {
			for (int j = - HALF_SIZE; j < HALF_SIZE + 1; ++j) {
				int newX = x + j;
				int newY = y + i;

				// Mat value is the pixel's normalized value
				unsigned char pixelValue = pixelPtr[newY * image.cols * cn + newX * cn];
				descriptorPtr[currentRow * NUM_NEIGHBOURS + currentColumn] = (pixelValue - min) / range;

				++currentColumn;
			}
		}
		++currentRow;
	}
}