#include "Tester.h"
#include "FeatureSystem.h"
#include "HARRIS.h"
#include "KLT.h"
#include "SIFT.h"
#include "SURF.h"
#include "MSER.h"
#include "opencv2/highgui/highgui.hpp"
#include "opencv2/imgproc/imgproc.hpp"
#include <iostream>
#include <map>

using namespace cv;
using namespace std;

// Macros/Defines
#define VERBOSE
#define OUTPUT_FOLDER "results\\"

// Uncomment if MSERs should be tested rather than the other feature systems
//#define TEST_MSER

#ifdef TEST_MSER
	// Change if images should be blurred prior to calculating MSERs
	#define TEST_MSER_BLUR false

	// Macro used to avoid repeating code unnecessarily
	#define redo_mser() contours1.clear(); contours2.clear(); \
	mser.findContours(iImage.imageMat, contours1); \
	mser.findContours(jImage.imageMat, contours2); \
	display.displayMSERs(iImage.imageMat, jImage.imageMat, contours1, contours2)
#endif

Tester::Tester()
: color(255)
#ifndef TEST_MSER
, writer(OUTPUT_FOLDER) 
#endif
{}

/**
 * Runs the test script on the given image files
 * Executes each feature system on each photo in a landmark against every other photo in that landmark
 */
void Tester::run(fileContainer& imageFiles) {
	// Initialize all the feature systems
	FeatureSystem* featureSystems[FeatureSystem::COUNT] = {
		new HarrisSystem(),
		new KLTSystem(),
		new SiftSystem(),
		new SurfSystem(),
	};

#ifdef TEST_MSER
	MserSystem mser(TEST_MSER_BLUR);
#endif

	fileContainer::const_iterator currentLandmark;

	map<string, ImageFile> imageFileObjs;		// file path to ImageFile object
	fileContainer::const_iterator start = imageFiles.begin();
	fileContainer::const_iterator end = imageFiles.end();

	// Optional skipping to a specific landmark
	const string startingLandmark = "";
	//const string startingLandmark = "colliseum";
	if (startingLandmark.empty() == false) {
		while (start->first.compare(startingLandmark) != 0) {
			++start;
			if (start == end) {
				// Non existant landmark name
				start = imageFiles.begin();
				break;
			}
		}
	}

	// Go through each landmark
	for (currentLandmark = start; currentLandmark != end; ++currentLandmark) {
		cout << "Landmark : " << currentLandmark->first << endl;
		cout << "===============================================" << endl;
		writer.setCurrentLandmark(currentLandmark->first);

		// Get the image names for the results writer		
		const fileList& files = currentLandmark->second;
		const size_t numFiles = files.size();
		const float numComparisons = (float) (numFiles * numFiles);
		std::vector<string> imageNames(numFiles);
		for (size_t i = 0; i < numFiles; ++i) {
			imageNames[i] = ImageFile::extractImageName(files[i]);
		}

#ifdef TEST_MSER
		// Unused variables, needed to avoid compile issues with conditional compiling
		int featureSystem = 0;
		FeatureSystem* currentSystem = NULL;
#else
		// Go through each feature system
		for (int featureSystem = 0; featureSystem < FeatureSystem::COUNT; ++featureSystem) {
			string systemName = FeatureSystem::getSystemName(featureSystem);
			cout << "Feature System: " << systemName << endl;
			FeatureSystem* currentSystem = featureSystems[featureSystem];			

			writer.setCurrentSystem(systemName);			
			writer.setImagesCompared(imageNames);
#endif
			int counter = 0;

			// Compare each image in the landmark to every other image in the landmark
			// Lazy loads each image, as well as their keypoints/descriptors (only loads them when necessary)
			for (size_t i = 0; i < numFiles; ++i) {
				// First image
				ImageFile& iImage = getImageObject(imageFileObjs, currentLandmark->first, files[i]);
				runSystemIfNecessary(currentSystem, featureSystem, iImage);
				writer.setCurrentImage(iImage.imageName);

				bool allowUndo = false;
				for (size_t j = 0; j < numFiles; ++j) {
					// Second image
					ImageFile& jImage = getImageObject(imageFileObjs, currentLandmark->first, files[j]);
					runSystemIfNecessary(currentSystem, featureSystem, jImage);

					cout << "Computing matches for " << iImage.getDisplayName() << " and " << jImage.getDisplayName() << endl;
					int result = 0;
#ifdef TEST_MSER
					// If testing MSER, only display both image's MSERs and don't record any results
					Mat blurI, blurJ;
					std::vector<std::vector<Point>> contours1, contours2;
					mser.findContours(iImage.imageMat, contours1);
					mser.findContours(jImage.imageMat, contours2);
					display.displayMSERs(iImage.imageMat, jImage.imageMat, contours1, contours2);
					do {
						result = waitKey(0);

						switch (result) {
						case 'c':
							// Display the original two images
							display.displayTwoImgs(iImage.imageMat, jImage.imageMat);
							break;
						case 'b':
							// Show the blurred versions of the images
							GaussianBlur(iImage.imageMat, blurI, Size(MserSystem::BLUR_SIZE,MserSystem::BLUR_SIZE), MserSystem::SIGMA);
							GaussianBlur(jImage.imageMat, blurJ, Size(MserSystem::BLUR_SIZE,MserSystem::BLUR_SIZE), MserSystem::SIGMA);
							display.displayTwoImgs(blurI, blurJ);
							break;
						case 'f':
							// Change whether a blur is done first, before getting the contours, or not
							mser.blurFirst = !mser.blurFirst;
							redo_mser();
							break;
						case 'm':
							// Display the MSERs of the two images
							display.displayMSERs(iImage.imageMat, jImage.imageMat, contours1, contours2);
							break;
						case ',':
							// Shrink the sigma value of the blur function
							MserSystem::SIGMA -= 0.5;
							if (MserSystem::SIGMA < 0.5) { MserSystem::SIGMA = 0.5; }
							redo_mser();
							break;
						case '.':
							// Increase the sigma value of the blur function
							MserSystem::SIGMA += 0.5;
							redo_mser();
							break;
						case ';':
							// Shrink the blur kernel size
							MserSystem::BLUR_SIZE -= 2;
							if (MserSystem::BLUR_SIZE < 5) { MserSystem::BLUR_SIZE = 5; }
							redo_mser();
							break;
						case '\'':
							// Increase the blur kernel size
							MserSystem::BLUR_SIZE += 2;
							redo_mser();
							break;
						}
					}
					while (result != 'y' && result != 'n');
					continue;		// Go to the next image comparison (skipping the code below)
#endif
					std::vector<DMatch> allMatches;		// Used only if they press 'm' to display all the matches, not just the good ones

					std::vector<Match> matches;
					currentSystem->flannFindMatches(iImage.keypoints[featureSystem], iImage.descriptors[featureSystem],
													jImage.keypoints[featureSystem], jImage.descriptors[featureSystem], matches);
					double median;
					std::vector<double> dispgrad;
					currentSystem->disparityGradientFilter(matches, dispgrad, &median);
					display.displayMatches(iImage.imageMat, jImage.imageMat, matches, dispgrad, median, currentSystem->disparityMedianScale, color);

#ifdef VERBOSE
					cout << "Done computing matches" << endl;
#endif
					bool goBack = false;
					do {
						result = waitKey(0);
						
						switch (result) {							
						case 'k':
							// Display all the keypoints
							display.displayKeypoints(iImage.imageMat, iImage.keypoints[featureSystem],
													 jImage.imageMat, jImage.keypoints[featureSystem], color);
							break;
						case 'c':
							// Display only the two images
							display.displayTwoImgs(iImage.imageMat, jImage.imageMat);
							break;
						case 'm':
							// Display all the matches
							if (allMatches.empty()) {
								// Lazy load
								FlannBasedMatcher matcher;
								matcher.match(iImage.descriptors[featureSystem], jImage.descriptors[featureSystem], allMatches);
							}
							display.displayMatches(iImage.imageMat, iImage.keypoints[featureSystem],
												   jImage.imageMat, jImage.keypoints[featureSystem],
												   allMatches);
							break;
						case 'f':
							// Display only the FLANN matching keypoints (ie: the more likely matches, this is the default)
							display.displayMatches(iImage.imageMat, jImage.imageMat, matches, color);
							break;
						case 'd':
							// Display only the matches that pass the disparity gradient
							display.displayMatches(iImage.imageMat, jImage.imageMat, matches, dispgrad, median, currentSystem->disparityMedianScale, color);
							break;
						case '.':
							color = (color == 255 ? 255 : color + 5);
							cout << color << endl;
							break;
						case ',':					
							color = (color == 0 ? 0 : color - 5);
							cout << color << endl;
							break;
						case 'u':
							// Undo the last result recorded.
							// Only allows you go to back one, not more. Can't go back if this is the first time the image is compared with another
							if (allowUndo) {
								writer.undoResult();
								goBack = true;
								allowUndo = false;		// Don't allow undo until we record a new result
								--counter;
							}
							break;
						}
					} while (result != 'y' && result != 'n' && goBack == false);

					if (goBack) {
						j -= 2;		// j will be incremented with continue
						// Go back to previous image
						continue;
					}

					// Append results
					if (result == 'y') {
						cout << ">> Recording a match" << endl;
						writer.recordResult(true);
					}
					else if (result == 'n') {
						cout << ">> Recording a mismatch" << endl;
						writer.recordResult(false);
					}
					// Recorded result, allow undo
					allowUndo = true;
					cout << "Completed " << ++counter << "/" << numComparisons << " (" << counter / numComparisons * 100 << "%)" << 
						" comparisons for this system" << endl;
				}
				// Next first image
				cout << endl;
				writer.finishImageComparison();
			}
#ifndef TEST_MSER
			// Next feature system
			writer.finishSystem();
		}
		// Next landmark, output current results
	}
#else
	}
#endif
	
	// Cleanup
	for (int i = 0; i < FeatureSystem::COUNT; ++i) {
		delete featureSystems[i];
	}
}

/**
 * Helper function to get an image object for the given landmark/fileName.
 * Initializes the ImageFile object if necessary
 */
ImageFile& Tester::getImageObject(map<string, ImageFile>& imageFileObjs, string landmarkName, string fileName) {
	ImageFile& image = imageFileObjs[fileName];
	if (image.isNULL()) {
		image.initialize(landmarkName, fileName);
		image.imageMat = imread(fileName, CV_LOAD_IMAGE_GRAYSCALE);
	}
	return image;
}

/**
 * If the given image hasn't had its keypoints/descriptors loaded, do so
 */
void Tester::runSystemIfNecessary(FeatureSystem* featureSystem, int systemIndex, ImageFile& image) {
	if (featureSystem == NULL) {
		return;
	}
	if (image.keypoints[systemIndex].empty()) {
#ifdef VERBOSE
		cout << "Calculating keypoints for " << image.getDisplayName() << endl;
#endif
		featureSystem->findFeatures(image.imageMat, image.keypoints[systemIndex]);
	}
	if (image.descriptors[systemIndex].total() == 0) {
#ifdef VERBOSE
		cout << "Calculating descriptors for " << image.getDisplayName() << endl;
#endif
		featureSystem->findDescriptors(image.imageMat, image.keypoints[systemIndex], image.descriptors[systemIndex]);
	}
}